//-----------------------------------------------------------------------------
// imu.cc : implements IMU class
//
//	- programmed by Boyoon Jung (boyoon@robotics.usc.edu)
//-----------------------------------------------------------------------------
#include "imu.h"
#include <math.h>

// header files for debugging info
#include <iostream>
using std::cerr;
using std::endl;


// constructor
MicroStrain3DMG::MicroStrain3DMG(char* device, tcflag_t baudrate)
{
    // open a serial port for communication to 3DM-G
    if (device)
	this->open(device, baudrate);
    else
	serial = 0;
}


// destructor
MicroStrain3DMG::~MicroStrain3DMG(void)
{
    // close the serial port
    this->close();
}


// open a serial port for communication to 3DM-G
bool MicroStrain3DMG::open(char* device, tcflag_t baudrate)
{
    // close a seiral port if there is one already opened
    if (serial)  {
	::tcsetattr(serial, TCSAFLUSH, &old_conf);
	::close(serial);
    }

    // open the serial port
    if ((serial=::open(device, O_RDWR|O_SYNC, S_IRUSR|S_IWUSR)) < 0) {
	::perror("cannot open the serial port");
	serial = 0;
	return false;
    }

    // save the current configuration of the serial port
    if (::tcgetattr(serial, &old_conf) < 0) {
	::perror("cannot retrieve terminal info.");
	::close(serial);
	serial = 0;
	return false;
    }

    // change the configuration of the serial port
    termios conf = old_conf;
    conf = old_conf;
    ::cfmakeraw(&conf);
    ::cfsetispeed(&conf, baudrate);
    ::cfsetospeed(&conf, baudrate);
    if (::tcsetattr(serial, TCSAFLUSH, &conf) < 0 || ::tcflush(serial, TCIOFLUSH) < 0) {
	::perror("cannot set terminal attributes.");
	::close(serial);
	serial = 0;
	return false;
    }

    // check if the communication is established by reading firmware info.
    cerr << "[Info] inializing IMU... ";
    if (readFirmwareVersion() == 0)
    {
	cerr << "MicroStrain3DMG::open(): fail to communicate with 3DM-G." << endl;
	::tcsetattr(serial, TCSAFLUSH, &old_conf);
	::close(serial);
	serial = 0;
	return false;
    }
    else
	cerr << "Done." << endl;

    // done
    return true;
}


// close the serial port
bool MicroStrain3DMG::close(void)
{
    if (serial) {
	// restore the old configuration
	::tcsetattr(serial, TCSAFLUSH, &old_conf);
	// close the device file
	::close(serial);
    }

    // done
    return true;
}


// read the instantaneously measured vectors
bool MicroStrain3DMG::read(imu_vector_t& data)
{
    // send the command for info retrieval
    if (! send(0x03, buffer, 23)) return false;

    // magnetometers
    data.mx = bytes2short(buffer+1) / 8192.f;
    data.my = bytes2short(buffer+3) / 8192.f;
    data.mz = bytes2short(buffer+5) / 8192.f;

    // accelerometers
    data.ax = bytes2short(buffer+7) / 8192.f;
    data.ay = bytes2short(buffer+9) / 8192.f;
    data.az = bytes2short(buffer+11) / 8192.f;

    // gyroscopes
    data.gx = bytes2short(buffer+13) / (GYRO_GAIN_SCALE * 8192.f * 0.0065536);
    data.gy = bytes2short(buffer+15) / (GYRO_GAIN_SCALE * 8192.f * 0.0065536);
    data.gz = bytes2short(buffer+17) / (GYRO_GAIN_SCALE * 8192.f * 0.0065536);

    // done
    return true;
}


// read the stablized vectors
bool MicroStrain3DMG::readStablized(imu_vector_t& data)
{
    // send the command for info retrieval
    if (! send(0x02, buffer, 23)) return false;

    // magnetometers
    data.mx = bytes2short(buffer+1) / 8192.f;
    data.my = bytes2short(buffer+3) / 8192.f;
    data.mz = bytes2short(buffer+5) / 8192.f;

    // accelerometers
    data.ax = bytes2short(buffer+7) / 8192.f;
    data.ay = bytes2short(buffer+9) / 8192.f;
    data.az = bytes2short(buffer+11) / 8192.f;

    // gyroscopes
    data.gx = bytes2short(buffer+13) / (GYRO_GAIN_SCALE * 8192.f * 0.0065536);
    data.gy = bytes2short(buffer+15) / (GYRO_GAIN_SCALE * 8192.f * 0.0065536);
    data.gz = bytes2short(buffer+17) / (GYRO_GAIN_SCALE * 8192.f * 0.0065536);

    // done
    return true;
}


// read the instantaneously computed transformation matrix
bool MicroStrain3DMG::read(imu_matrix_t& data)
{
    // send the command for info retrieval
    if (! send(0x0A, buffer, 23)) return false;

    // transformation matrix
    data.m00 = bytes2short(buffer+1) / 8192.f;
    data.m10 = bytes2short(buffer+3) / 8192.f;
    data.m20 = bytes2short(buffer+5) / 8192.f;
    data.m01 = bytes2short(buffer+7) / 8192.f;
    data.m11 = bytes2short(buffer+9) / 8192.f;
    data.m21 = bytes2short(buffer+11) / 8192.f;
    data.m02 = bytes2short(buffer+13) / 8192.f;
    data.m12 = bytes2short(buffer+15) / 8192.f;
    data.m22 = bytes2short(buffer+17) / 8192.f;

    // normalization
    double det =  data.m00 * data.m11 * data.m22
		+ data.m01 * data.m12 * data.m20
		+ data.m02 * data.m10 * data.m21
		- data.m02 * data.m11 * data.m20
		- data.m12 * data.m21 * data.m00
		- data.m22 * data.m01 * data.m10;
    data.m00 /= det;
    data.m01 /= det;
    data.m02 /= det;
    data.m10 /= det;
    data.m11 /= det;
    data.m12 /= det;
    data.m20 /= det;
    data.m21 /= det;
    data.m22 /= det;

    // done
    return true;
}


// read the gyroscopically stablized transformation matrix
bool MicroStrain3DMG::readStablized(imu_matrix_t& data)
{
    // send the command for info retrieval
    if (! send(0x0B, buffer, 23)) return false;

    // transformation matrix
    data.m00 = bytes2short(buffer+1) / 8192.f;
    data.m10 = bytes2short(buffer+3) / 8192.f;
    data.m20 = bytes2short(buffer+5) / 8192.f;
    data.m01 = bytes2short(buffer+7) / 8192.f;
    data.m11 = bytes2short(buffer+9) / 8192.f;
    data.m21 = bytes2short(buffer+11) / 8192.f;
    data.m02 = bytes2short(buffer+13) / 8192.f;
    data.m12 = bytes2short(buffer+15) / 8192.f;
    data.m22 = bytes2short(buffer+17) / 8192.f;

    // normalization
    double det =  data.m00 * data.m11 * data.m22
		+ data.m01 * data.m12 * data.m20
		+ data.m02 * data.m10 * data.m21
		- data.m02 * data.m11 * data.m20
		- data.m12 * data.m21 * data.m00
		- data.m22 * data.m01 * data.m10;
    data.m00 /= det;
    data.m01 /= det;
    data.m02 /= det;
    data.m10 /= det;
    data.m11 /= det;
    data.m12 /= det;
    data.m20 /= det;
    data.m21 /= det;
    data.m22 /= det;

    // done
    return true;
}


// read the instantaneously computed quaternion
bool MicroStrain3DMG::read(imu_quaternion_t& data)
{
    // send the command for info retrieval
    if (! send(0x04, buffer, 13)) return false;

    // Quaternion
    data.q0 = bytes2short(buffer+1) / 8192.f;
    data.q1 = bytes2short(buffer+3) / 8192.f;
    data.q2 = bytes2short(buffer+5) / 8192.f;
    data.q3 = bytes2short(buffer+7) / 8192.f;

    // done
    return true;
}


// read the gyroscopically stablized quaternion
bool MicroStrain3DMG::readStablized(imu_quaternion_t& data)
{
    // send the command for info retrieval
    if (! send(0x05, buffer, 13)) return false;

    // Quaternion
    data.q0 = bytes2short(buffer+1) / 8192.f;
    data.q1 = bytes2short(buffer+3) / 8192.f;
    data.q2 = bytes2short(buffer+5) / 8192.f;
    data.q3 = bytes2short(buffer+7) / 8192.f;

    // done
    return true;
}


// read the instantaneously computed Euler angles
bool MicroStrain3DMG::read(imu_euler_t& data)
{
    /*
     * This command is defined in the manual, but it does NOT work.
     * It seems like that 3DM-G simply ignore the command and return none.
     * This is a simple hack to compute Euler angles from a transformation matrix.
     */

    /*
    // send the command for info retrieval
    if (! send(0x0D, buffer, 11)) return false;

    // Euler angles
    data.roll = bytes2short(buffer+1) * 360 / 65536.f;
    data.pitch = bytes2short(buffer+3) * 360 / 65536.f;
    data.yaw = bytes2short(buffer+5) * 360 / 65536.f;
    */

    // read a transformation matrix
    static imu_matrix_t matrix;
    this->read(matrix);

    // compute a roll angle from the transformation matrix
    data.roll = atan2(matrix.m12, matrix.m22);

    // compute a roll angle from the transformation matrix
    double det = -matrix.m02 * matrix.m02 + 1;
    if (det <= 0) {
	if (matrix.m02 > 0) data.pitch = -M_PI / 2;
	else data.pitch = M_PI / 2;
    }
    else
	data.pitch = -atan(matrix.m02 / (det*det));

    // compute a yaw angle from the transformation matrix
    data.yaw = atan2(matrix.m01, matrix.m00);

    // done
    return true;
}


// read the gyroscopically stablized Euler angles
bool MicroStrain3DMG::readStablized(imu_euler_t& data)
{
    /*
     * This command is defined in the manual, but it does NOT work.
     * It seems like that 3DM-G simply ignore the command and return none.
     * This is a simple hack to compute Euler angles from a transformation matrix.
     */

    /*
    // send the command for info retrieval
    if (! send(0x0E, buffer, 11)) return false;

    // Euler angles
    data.roll = bytes2short(buffer+1) * 360 / 65536.f;
    data.pitch = bytes2short(buffer+3) * 360 / 65536.f;
    data.yaw = bytes2short(buffer+5) * 360 / 65536.f;
    */

    // read a transformation matrix
    static imu_matrix_t matrix;
    this->readStablized(matrix);

    // compute a roll angle from the transformation matrix
    data.roll = atan2(matrix.m12, matrix.m22);

    // compute a roll angle from the transformation matrix
    double det = -matrix.m02 * matrix.m02 + 1;
    if (det <= 0) {
	if (matrix.m02 > 0) data.pitch = -M_PI / 2;
	else data.pitch = M_PI / 2;
    }
    else
	data.pitch = -atan(matrix.m02 / (det*det));

    // compute a yaw angle from the transformation matrix
    data.yaw = atan2(matrix.m01, matrix.m00);

    // done
    return true;
}


// read the gyroscopically stablized quaternion & vectors
bool MicroStrain3DMG::read(imu_quaternion_t& q, imu_vector_t& v)
{
    // send the command for info retrieval
    if (! send(0x0C, buffer, 31)) return false;

    // Quaternion
    q.q0 = bytes2short(buffer+1) / 8192.f;
    q.q1 = bytes2short(buffer+3) / 8192.f;
    q.q2 = bytes2short(buffer+5) / 8192.f;
    q.q3 = bytes2short(buffer+7) / 8192.f;

    // magnetometers
    v.mx = bytes2short(buffer+9) / 8192.f;
    v.my = bytes2short(buffer+11) / 8192.f;
    v.mz = bytes2short(buffer+13) / 8192.f;

    // accelerometers
    v.ax = bytes2short(buffer+15) / 8192.f;
    v.ay = bytes2short(buffer+17) / 8192.f;
    v.az = bytes2short(buffer+19) / 8192.f;

    // gyroscopes
    v.gx = bytes2short(buffer+21) / (GYRO_GAIN_SCALE * 8192.f * 0.0065536);
    v.gy = bytes2short(buffer+23) / (GYRO_GAIN_SCALE * 8192.f * 0.0065536);
    v.gz = bytes2short(buffer+25) / (GYRO_GAIN_SCALE * 8192.f * 0.0065536);

    // done
    return true;
}


// read the raw voltage outputs
bool MicroStrain3DMG::read(imu_voltage_t& data)
{
    // send the command for info retrieval
    if (! send(0x01, buffer, 23)) return false;

    // magnetometers
    data.mx = (unsigned short) bytes2short(buffer+1);
    data.my = (unsigned short) bytes2short(buffer+3);
    data.mz = (unsigned short) bytes2short(buffer+5);

    // accelerometers
    data.ax = (unsigned short) bytes2short(buffer+7);
    data.ay = (unsigned short) bytes2short(buffer+9);
    data.az = (unsigned short) bytes2short(buffer+11);

    // gyroscopes
    data.gx = (unsigned short) bytes2short(buffer+13);
    data.gy = (unsigned short) bytes2short(buffer+15);
    data.gz = (unsigned short) bytes2short(buffer+17);

    // done
    return true;
}


// read the temperature of the interior of the 3DM-G unit
double MicroStrain3DMG::readTemperature(void)
{
    // send the command for info retrieval
    if (! send(0x07, buffer, 7)) return false;

    // return the temparature
    return bytes2short(buffer+1) * 0.12207f;
}


// capture the gyroscope bias
bool MicroStrain3DMG::captureGyroBias(void)
{
    // send the command for info retrieval
    if (! send(0x06, buffer, 5)) return false;

    // done
    return true;
}


// read the version number of firmware
int MicroStrain3DMG::readFirmwareVersion(void)
{
    // send the command for info retrieval
    if (! send(0xF0, buffer, 5)) return 0;

    // decode the information
    return (unsigned short) bytes2short(buffer+1);
}


// read the device serial number
int MicroStrain3DMG::readSerialNumber(void)
{
    // send the command for info retrieval
    if (! send(0xF1, buffer, 5)) return 0;

    // decode the information
    return (unsigned short) bytes2short(buffer+1);
}


