//----------------------------------------------------------------------------
// qimu.cc : implementation of IMU visualization classes
//
// 	- programmed by Boyoon Jung (boyoon@robotics.usc.edu)
//----------------------------------------------------------------------------
#include "qimu.h"
#include <math.h>

// header files for Qt library
#include <qtabwidget.h>
#include <qpushbutton.h>
#include <qslider.h>
#include <qcombobox.h>
#include <qlabel.h>
#include <qcursor.h>
#include <qcompass.h>


// constructor
Q3dmg::Q3dmg(MicroStrain3DMG* imu, QWidget* parent, const char* name, WFlags f)
	: Q3dmgUI(parent, name, f)
{
    // store the pointer of the IMU
    if (imu) setIMU(imu);
    else imu = 0;

    // configure dial widgets
    rollDial->drawEWSN(false);
    rollDial->drawBidirectional(false);
    pitchDial->drawEWSN(false);
    pitchDial->drawBidirectional(false);
    yawDial->drawEWSN(false);
    yawDial->drawBidirectional(false);

    // connect signals and slots
    connect(updateButton, SIGNAL(stateChanged(int)), SLOT(updateSlot(int)));
    connect(captureButton, SIGNAL(clicked()), SLOT(captureSlot()));
}


// destructor
Q3dmg::~Q3dmg(void)
{
    // free the allocated objects
}


// set the IMU
void Q3dmg::setIMU(MicroStrain3DMG* imu)
{
    static char tmpstr[16];

    // store the pointer of the IMU
    this->imu = imu;

    // retrieve the basic information
    int firmware = imu->readFirmwareVersion();
    int serial = imu->readSerialNumber();

    // update the basic information
    sprintf(tmpstr, "%d.%d.%d", firmware/1000, (firmware%1000)/10, firmware%10);
    fvText->setText(tmpstr);
    snText->setNum(serial);
}


// timer for periodical update
void Q3dmg::timerEvent(QTimerEvent*)
{
    static int index = -1;
    // update the apporpriate information
    switch (mainTab->currentPageIndex())
    {
	case 0: updatePose(); break;
	case 1: updateOrientation(); break;
	case 2: updateVector(); break;
	case 3: updateVoltage(); break;
	case 4: updateMisc(); break;
    }
}


// update the pose estimation page
void Q3dmg::updatePose(void)
{
    static imu_quaternion_t quaternion;

    if (imu)
    {
	// retrieve data
	if (pStabCombo->currentItem())
	    imu->read(quaternion);
	else
	    imu->readStablized(quaternion);

	// update OpenGL visualization
	pose->setValue(quaternion.q0, quaternion.q1, quaternion.q2, quaternion.q3);
    }
}


// update the orientation page
void Q3dmg::updateOrientation(void)
{
    static imu_matrix_t matrix;
    static imu_quaternion_t quaternion;
    static imu_euler_t euler;

    if (imu)
    {
	// retrieve data
	if (oStabCombo->currentItem()) {
	    imu->read(euler);
	    imu->read(matrix);
	    imu->read(quaternion);
	}
	else {
	    imu->readStablized(euler);
	    imu->readStablized(matrix);
	    imu->readStablized(quaternion);
	}

	// update the Euler angles
	rollDial->setValue(euler.roll*180/M_PI);
	pitchDial->setValue(euler.pitch*180/M_PI);
	yawDial->setValue(euler.yaw*180/M_PI);

	// update the quaternion
	q0->setNum(quaternion.q0);
	q1->setNum(quaternion.q1);
	q2->setNum(quaternion.q2);
	q3->setNum(quaternion.q3);

	// update the transformation matrix
	tm00->setNum(matrix.m00); tm01->setNum(matrix.m01); tm02->setNum(matrix.m02);
	tm10->setNum(matrix.m10); tm11->setNum(matrix.m11); tm12->setNum(matrix.m12);
	tm20->setNum(matrix.m20); tm21->setNum(matrix.m21); tm22->setNum(matrix.m22);
    }
}


// update the vector page
void Q3dmg::updateVector(void)
{
    static imu_vector_t vector;

    if (imu)
    {
	// retrieve data
	if (vStabCombo->currentItem())
	    imu->read(vector);
	else
	    imu->readStablized(vector);

	// update all vectors
	vmx->setNum(vector.mx); vmy->setNum(vector.my); vmz->setNum(vector.mz);
	vax->setNum(vector.ax); vay->setNum(vector.ay); vaz->setNum(vector.az);
	vgx->setNum(vector.gx); vgy->setNum(vector.gy); vgz->setNum(vector.gz);
    }
}


// update the voltage page
void Q3dmg::updateVoltage(void)
{
    static imu_voltage_t voltage;

    if (imu)
    {
	// retrive data
	imu->read(voltage);

	// update all voltage levels
	mxSlider->setValue(4096-voltage.mx);
	mySlider->setValue(4096-voltage.my);
	mzSlider->setValue(4096-voltage.mz);

	axSlider->setValue(4096-voltage.ax);
	aySlider->setValue(4096-voltage.ay);
	azSlider->setValue(4096-voltage.az);

	gxSlider->setValue(4096-voltage.gx);
	gySlider->setValue(4096-voltage.gy);
	gzSlider->setValue(4096-voltage.gz);
    }
}


// update the misc page
void Q3dmg::updateMisc(void)
{
    static double temperature;
    if (imu)
    {
	// retrive data
	temperature = imu->readTemperature();

	// update temperature
	tempSlider->setValue(int(100-temperature+0.5));
	tempText->setNum(temperature);
    }
}


// start/stop updating IMU information
void Q3dmg::updateSlot(int state)
{
    if (state == QButton::On) startTimer(100);
    else killTimers();
}


// capture gyroscope bias
void Q3dmg::captureSlot(void)
{
    if (imu) {
	captureButton->setCursor(WaitCursor);
	imu->captureGyroBias();
	captureButton->unsetCursor();
    }
}
